@@ -3,18 +3,18 @@
 # License-Filename: LICENSES/GPL-3.0
 
 import logging
-import os.path
 
-from gi.repository import GLib, Gtk, Gdk, Gio, Pango
+from gi.repository import GLib, GObject, Gtk, Gio, Pango
 
 from gtweak.tweakmodel import Tweak, TweakGroup
 from gtweak.gsettings import GSettingsSetting, GSettingsFakeSetting, GSettingsMissingError
-from gtweak.gtksettings import GtkSettingsManager
 from gtweak.gshellwrapper import GnomeShellFactory
+from gtweak.utils import SchemaList
 
 UI_BOX_SPACING = 4
 _shell = GnomeShellFactory().get_shell()
 
+
 def build_label_beside_widget(txt, *widget, **kwargs):
     """
     Builds a HBox containing widgets.
@@ -44,7 +44,7 @@ def build_label_beside_widget(txt, *widget, **kwargs):
 
     hbox.props.spacing = UI_BOX_SPACING
     lbl = Gtk.Label(label=txt)
-    lbl.props.ellipsize = Pango.EllipsizeMode.END
+    lbl.props.ellipsize = Pango.EllipsizeMode.NONE
     lbl.props.xalign = 0.0
     lbl.set_has_tooltip(True)
     lbl.connect("query-tooltip", show_tooltip_when_ellipsized)
@@ -62,13 +62,14 @@ def build_label_beside_widget(txt, *widget, **kwargs):
     for w in widget:
         hbox.pack_start(w, False, False, 0)
 
-    #For Atk, indicate that the rightmost widget, usually the switch relates to the
-    #label. By convention this is true in the great majority of cases. Settings that
-    #construct their own widgets will need to set this themselves
+    # For Atk, indicate that the rightmost widget, usually the switch relates to the
+    # label. By convention this is true in the great majority of cases. Settings that
+    # construct their own widgets will need to set this themselves
     lbl.set_mnemonic_widget(widget[-1])
 
     return hbox
 
+
 def build_combo_box_text(selected, *values):
     """
     builds a GtkComboBox and model containing the supplied values.
@@ -79,7 +80,7 @@ def build_combo_box_text(selected, *values):
 
     selected_iter = None
     for (val, name) in values:
-        _iter = store.append( (val, name) )
+        _iter = store.append((val, name))
         if val == selected:
             selected_iter = _iter
 
@@ -92,24 +93,26 @@ def build_combo_box_text(selected, *values):
 
     return combo
 
+
 def build_horizontal_sizegroup():
     sg = Gtk.SizeGroup(mode=Gtk.SizeGroupMode.HORIZONTAL)
     sg.props.ignore_hidden = True
     return sg
 
+
 def build_tight_button(stock_id):
     button = Gtk.Button()
     button.set_relief(Gtk.ReliefStyle.NONE)
     button.set_focus_on_click(False)
     button.add(Gtk.Image.new_from_stock(stock_id, Gtk.IconSize.MENU))
-    data =  ".button {\n" \
-            "-GtkButton-default-border : 0px;\n" \
-            "-GtkButton-default-outside-border : 0px;\n" \
-            "-GtkButton-inner-border: 0px;\n" \
-            "-GtkWidget-focus-line-width : 0px;\n" \
-            "-GtkWidget-focus-padding : 0px;\n" \
-            "padding: 0px;\n" \
-            "}"
+    data = ".button {\n" \
+           "-GtkButton-default-border : 0px;\n" \
+           "-GtkButton-default-outside-border : 0px;\n" \
+           "-GtkButton-inner-border: 0px;\n" \
+           "-GtkWidget-focus-line-width : 0px;\n" \
+           "-GtkWidget-focus-padding : 0px;\n" \
+           "padding: 0px;\n" \
+           "}"
     provider = Gtk.CssProvider()
     provider.load_from_data(data)
     # 600 = GTK_STYLE_PROVIDER_PRIORITY_APPLICATION
@@ -118,26 +121,31 @@ def build_tight_button(stock_id):
 
 
 class _GSettingsTweak(Tweak):
-    def __init__(self, name, schema_name, key_name, **options):
+    def __init__(self, title, schema_name, key_name, schema_dir=None, resettable_to_default=True, loaded=True, **options):
         self.schema_name = schema_name
         self.key_name = key_name
         self._extra_info = None
+
+        if resettable_to_default:
+            SchemaList.insert(key_name, schema_name, schema_dir)
+
         if 'uid' not in options:
             options['uid'] = key_name
         try:
             self.settings = GSettingsSetting(schema_name, **options)
             Tweak.__init__(self,
-                name,
-                options.get("description",self.settings.schema_get_description(key_name)),
+                title,
+                options.get("description", self.settings.schema_get_description(key_name)),
+                loaded=loaded,
                 **options)
         except GSettingsMissingError as e:
             self.settings = GSettingsFakeSetting()
-            Tweak.__init__(self,"","")
+            Tweak.__init__(self, "", "")
             self.loaded = False
             logging.info("GSetting missing %s", e)
         except KeyError:
             self.settings = GSettingsFakeSetting()
-            Tweak.__init__(self,"","")
+            Tweak.__init__(self, "", "")
             self.loaded = False
             logging.info("GSettings missing key %s (key %s)" % (schema_name, key_name))
 
@@ -153,13 +161,14 @@ class _GSettingsTweak(Tweak):
             self._extra_info = self.settings.schema_get_summary(self.key_name)
         return self._extra_info
 
+
 class _DependableMixin(object):
 
     def add_dependency_on_tweak(self, depends, depends_how):
         if isinstance(depends, Tweak):
             self._depends = depends
             if depends_how is None:
-                depends_how = lambda x,kn: x.get_boolean(kn)
+                depends_how = lambda x, kn: x.get_boolean(kn)
             self._depends_how = depends_how
 
             sensitive = self._depends_how(
@@ -171,34 +180,47 @@ class _DependableMixin(object):
             depends.settings.connect("changed::%s" % depends.key_name, self._on_changed_depend)
 
     def _on_changed_depend(self, settings, key_name):
-        sensitive = self._depends_how(settings,key_name)
+        sensitive = self._depends_how(settings, key_name)
         self.set_sensitive(sensitive)
 
+class TweakListBoxRow(Gtk.ListBoxRow):
+    __gtype_name__ = "GTweakListBoxRow"
+
+    tweakname = GObject.Property(type=str)
+
 class ListBoxTweakGroup(Gtk.ListBox, TweakGroup):
-    def __init__(self, name, *tweaks, **options):
+    def __init__(self, name, title, *tweaks, **options):
         if 'uid' not in options:
             options['uid'] = self.__class__.__name__
+        if 'activatable' not in options:
+            activatable = False
+        else:
+            activatable = options['activatable']
         Gtk.ListBox.__init__(self,
                         selection_mode=Gtk.SelectionMode.NONE,
                         name=options['uid'])
+        
         self.get_style_context().add_class(
-                        options.get('css_class','tweak-group'))
+                        options.get('css_class', 'tweak-group'))
+
         self.props.margin = 20
         self.props.vexpand = False
         self.props.valign = Gtk.Align.START
+        self.props.hexpand = True
+        self.props.halign = Gtk.Align.FILL
 
         self._sg = Gtk.SizeGroup(mode=Gtk.SizeGroupMode.HORIZONTAL)
         self._sg.props.ignore_hidden = True
 
-        TweakGroup.__init__(self, name, **options)
+        TweakGroup.__init__(self, name, title, **options)
 
         for t in tweaks:
-            self.add_tweak_row(t)
+            self.add_tweak_row(t, activatable)
 
-    #FIXME: need to add remove_tweak_row and remove_tweak (which clears
-    #the search cache etc)
+    # FIXME: need to add remove_tweak_row and remove_tweak (which clears
+    # the search cache etc)
 
-    def add_tweak_row(self, t, position=None):
+    def add_tweak_row(self, t, activatable=False, position=None):
         if self.add_tweak(t):
             if isinstance(t, Gtk.ListBoxRow):
                 row = t
@@ -208,6 +230,7 @@ class ListBoxTweakGroup(Gtk.ListBox, TweakGroup):
                 if isinstance(t, Title):
                     row.get_style_context().add_class("title")
                 row.add(t)
+            row.set_activatable(activatable)
             if position is None:
                 self.add(row)
             else:
@@ -216,10 +239,11 @@ class ListBoxTweakGroup(Gtk.ListBox, TweakGroup):
                 self._sg.add_widget(t.widget_for_size_group)
             return row
 
+
 class GSettingsCheckTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
-    def __init__(self, name, schema_name, key_name, **options):
+    def __init__(self, name, title, schema_name, key_name, **options):
         Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
-        _GSettingsTweak.__init__(self, name, schema_name, key_name, **options)
+        _GSettingsTweak.__init__(self, title, schema_name, key_name, **options)
 
         widget = Gtk.CheckButton.new_with_label(name)
         self.settings.bind(
@@ -234,10 +258,11 @@ class GSettingsCheckTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
                 options.get("depends_how")
         )
 
+
 class GSettingsSwitchTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
-    def __init__(self, name, schema_name, key_name, **options):
+    def __init__(self, title, schema_name, key_name, **options):
         Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
-        _GSettingsTweak.__init__(self, name, schema_name, key_name, **options)
+        _GSettingsTweak.__init__(self, title, schema_name, key_name, **options)
 
         w = Gtk.Switch()
         self.settings.bind(key_name, w, "active", Gio.SettingsBindFlags.DEFAULT)
@@ -249,8 +274,8 @@ class GSettingsSwitchTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
 
         vbox1 = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
         vbox1.props.spacing = UI_BOX_SPACING
-        lbl = Gtk.Label(label=name)
-        lbl.props.ellipsize = Pango.EllipsizeMode.END
+        lbl = Gtk.Label(label=title)
+        lbl.props.ellipsize = Pango.EllipsizeMode.NONE
         lbl.props.xalign = 0.0
         vbox1.pack_start(lbl, True, True, 0)
 
@@ -274,37 +299,40 @@ class GSettingsSwitchTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
         self.pack_start(vbox2, False, False, 0)
         self.widget_for_size_group = None
 
+
 class GSettingsFontButtonTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
-    def __init__(self, name, schema_name, key_name, **options):
+    def __init__(self, title, schema_name, key_name, **options):
         Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
-        _GSettingsTweak.__init__(self, name, schema_name, key_name, **options)
+        _GSettingsTweak.__init__(self, title, schema_name, key_name, **options)
 
         w = Gtk.FontButton()
         w.set_use_font(True)
         self.settings.bind(key_name, w, "font-name", Gio.SettingsBindFlags.DEFAULT)
-        build_label_beside_widget(name, w, hbox=self)
+        build_label_beside_widget(title, w, hbox=self)
         self.widget_for_size_group = w
 
-class GSettingsRangeTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
-    def __init__(self, name, schema_name, key_name, **options):
-        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
-        _GSettingsTweak.__init__(self, name, schema_name, key_name, **options)
 
-        #returned variant is range:(min, max)
+class GSettingsRangeTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
+    def __init__(self, title, schema_name, key_name, **options):
+        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
+        _GSettingsTweak.__init__(self, title, schema_name, key_name, **options)
+
+        # returned variant is range:(min, max)
         _min, _max = self.settings.get_range(key_name)[1]
 
         w = Gtk.HScale.new_with_range(_min, _max, options.get('adjustment_step', 1))
         self.settings.bind(key_name, w.get_adjustment(), "value", Gio.SettingsBindFlags.DEFAULT)
 
-        build_label_beside_widget(self.name, w, hbox=self)
+        build_label_beside_widget(self.title, w, hbox=self)
         self.widget_for_size_group = w
 
-class GSettingsSpinButtonTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
-    def __init__(self, name, schema_name, key_name, **options):
-        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
-        _GSettingsTweak.__init__(self, name, schema_name, key_name, **options)
 
-        #returned variant is range:(min, max)
+class GSettingsSpinButtonTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
+    def __init__(self, title, schema_name, key_name, **options):
+        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
+        _GSettingsTweak.__init__(self, title, schema_name, key_name, **options)
+
+        # returned variant is range:(min, max)
         _min, _max = self.settings.get_range(key_name)[1]
 
         adjustment = Gtk.Adjustment(value=0, lower=_min, upper=_max, step_increment=options.get('adjustment_step', 1))
@@ -313,7 +341,7 @@ class GSettingsSpinButtonTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
         w.set_digits(options.get('digits', 0))
         self.settings.bind(key_name, adjustment, "value", Gio.SettingsBindFlags.DEFAULT)
 
-        build_label_beside_widget(name, w, hbox=self)
+        build_label_beside_widget(title, w, hbox=self)
         self.widget_for_size_group = w
 
         self.add_dependency_on_tweak(
@@ -321,25 +349,26 @@ class GSettingsSpinButtonTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
                 options.get("depends_how")
         )
 
+
 class GSettingsComboEnumTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
-    def __init__(self, name, schema_name, key_name, **options):
+    def __init__(self, title, schema_name, key_name, **options):
         Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
-        _GSettingsTweak.__init__(self, name, schema_name, key_name, **options)
+        _GSettingsTweak.__init__(self, title, schema_name, key_name, **options)
 
         _type, values = self.settings.get_range(key_name)
         value = self.settings.get_string(key_name)
         self.settings.connect('changed::'+self.key_name, self._on_setting_changed)
 
-        w = build_combo_box_text(value, *[(v,v.replace("-"," ").title()) for v in values])
+        w = build_combo_box_text(value, *[(v, v.replace("-", " ").title()) for v in values])
         w.connect('changed', self._on_combo_changed)
         self.combo = w
 
-        build_label_beside_widget(name, w, hbox=self)
+        build_label_beside_widget(title, w, hbox=self)
         self.widget_for_size_group = w
 
     def _values_are_different(self):
-        #to stop bouncing back and forth between changed signals. I suspect there must be a nicer
-        #Gio.settings_bind way to fix this
+        # to stop bouncing back and forth between changed signals. I suspect there must be a nicer
+        # Gio.settings_bind way to fix this
         return self.settings.get_string(self.key_name) != \
                self.combo.get_model().get_value(self.combo.get_active_iter(), 0)
 
@@ -357,13 +386,14 @@ class GSettingsComboEnumTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
         if self._values_are_different():
             self.settings.set_string(self.key_name, val)
 
-class GSettingsComboTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
-    def __init__(self, name, schema_name, key_name, key_options, **options):
-        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
-        _GSettingsTweak.__init__(self, name, schema_name, key_name, **options)
 
-        #check key_options is iterable
-        #and if supplied, check it is a list of 2-tuples
+class GSettingsComboTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
+    def __init__(self, title, schema_name, key_name, key_options, **options):
+        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
+        _GSettingsTweak.__init__(self, title, schema_name, key_name, **options)
+
+        # check key_options is iterable
+        # and if supplied, check it is a list of 2-tuples
         assert len(key_options) >= 0
         if len(key_options):
             assert len(key_options[0]) == 2
@@ -375,7 +405,7 @@ class GSettingsComboTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
         self.combo.connect('changed', self._on_combo_changed)
         self.settings.connect('changed::'+self.key_name, self._on_setting_changed)
 
-        build_label_beside_widget(name, self.combo,hbox=self)
+        build_label_beside_widget(title, self.combo, hbox=self)
         self.widget_for_size_group = self.combo
 
     def _on_setting_changed(self, setting, key):
@@ -398,10 +428,11 @@ class GSettingsComboTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
     @property
     def extra_info(self):
         if self._extra_info is None:
-           self._extra_info = self.settings.schema_get_summary(self.key_name)
-           self._extra_info += " " + " ".join(op[0] for op in self._key_options)
+            self._extra_info = self.settings.schema_get_summary(self.key_name)
+            self._extra_info += " " + " ".join(op[0] for op in self._key_options)
         return self._extra_info
 
+
 class FileChooserButton(Gtk.FileChooserButton):
     def __init__(self, title, local_only, mimetypes):
         Gtk.FileChooserButton.__init__(self, title=title)
@@ -412,22 +443,23 @@ class FileChooserButton(Gtk.FileChooserButton):
                 f.add_mime_type(m)
             self.set_filter(f)
 
-        #self.set_width_chars(15)
+        # self.set_width_chars(15)
         self.set_local_only(local_only)
         self.set_action(Gtk.FileChooserAction.OPEN)
 
+
 class GSettingsFileChooserButtonTweak(Gtk.Box, _GSettingsTweak, _DependableMixin):
-    def __init__(self, name, schema_name, key_name, local_only, mimetypes, **options):
+    def __init__(self, title, schema_name, key_name, local_only, mimetypes, **options):
         Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
-        _GSettingsTweak.__init__(self, name, schema_name, key_name, **options)
+        _GSettingsTweak.__init__(self, title, schema_name, key_name, **options)
 
         self.settings.connect('changed::'+self.key_name, self._on_setting_changed)
 
-        self.filechooser = FileChooserButton(name,local_only,mimetypes)
+        self.filechooser = FileChooserButton(title, local_only, mimetypes)
         self.filechooser.set_uri(self.settings.get_string(self.key_name))
         self.filechooser.connect("file-set", self._on_file_set)
 
-        build_label_beside_widget(name, self.filechooser, hbox=self)
+        build_label_beside_widget(title, self.filechooser, hbox=self)
         self.widget_for_size_group = self.filechooser
 
     def _values_are_different(self):
@@ -441,16 +473,17 @@ class GSettingsFileChooserButtonTweak(Gtk.Box, _GSettingsTweak, _DependableMixin
         if uri and self._values_are_different():
             self.settings.set_string(self.key_name, uri)
 
+
 class GetterSetterSwitchTweak(Gtk.Box, Tweak):
-    def __init__(self, name, **options):
+    def __init__(self, title, **options):
         Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
-        Tweak.__init__(self, name, options.get("description",""), **options)
+        Tweak.__init__(self, title, options.get("description", ""), **options)
 
         sw = Gtk.Switch()
         sw.set_active(self.get_active())
         sw.connect("notify::active", self._on_toggled)
 
-        build_label_beside_widget(name, sw, hbox=self)
+        build_label_beside_widget(title, sw, hbox=self)
 
     def _on_toggled(self, sw, pspec):
         self.set_active(sw.get_active())
@@ -461,8 +494,9 @@ class GetterSetterSwitchTweak(Gtk.Box, Tweak):
     def set_active(self, v):
         raise NotImplementedError()
 
+
 class Title(Gtk.Box, Tweak):
-    def __init__(self, name, desc, **options):
+    def __init__(self, name,  desc, **options):
         Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
         Tweak.__init__(self, name, desc, **options)
         widget = Gtk.Label()
@@ -472,11 +506,12 @@ class Title(Gtk.Box, Tweak):
             widget.set_margin_top(10)
         self.add(widget)
 
+
 class GSettingsSwitchTweakValue(Gtk.Box, _GSettingsTweak):
 
-    def __init__(self, name, schema_name, key_name, **options):
+    def __init__(self, title, schema_name, key_name, **options):
         Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
-        _GSettingsTweak.__init__(self, name, schema_name, key_name, **options)
+        _GSettingsTweak.__init__(self, title, schema_name, key_name, **options)
 
         sw = Gtk.Switch()
         sw.set_active(self.get_active())
@@ -484,8 +519,8 @@ class GSettingsSwitchTweakValue(Gtk.Box, _GSettingsTweak):
 
         vbox1 = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
         vbox1.props.spacing = UI_BOX_SPACING
-        lbl = Gtk.Label(label=name)
-        lbl.props.ellipsize = Pango.EllipsizeMode.END
+        lbl = Gtk.Label(label=title)
+        lbl.props.ellipsize = Pango.EllipsizeMode.NONE
         lbl.props.xalign = 0.0
         vbox1.pack_start(lbl, True, True, 0)
 
